'use strict'
import os from 'os'
import crypto from 'crypto'
import redis from 'redis'
import kafka from 'kafka-node'
import express from 'express'
import uuidv4 from 'uuid/v4'
import Worker from './worker.js'
import Fork from './fork.js'

class Coppel {
  constructor (config) {
    this._hostname = os.hostname()
    config.KAFKA = config.KAFKA.split('2181').join('9092')

    this._config = config
    this._config.MS_REDIS = process.env.MS_REDIS || null
    this._config.TOPIC = `${config.APPLICATION_APP}_${config.APPLICATION_VERSION}_${config.APPLICATION_NAME}`
    if (config.APPLICATION_APP === 'appcoppel' && config.APPLICATION_VERSION === 'v1') {
      this._config.TOPIC = config.APPLICATION_NAME
    }

    this._config.DEBUG = this._config.DEBUG && this._config.DEBUG.toString() === 'true'
    this._config.APPLICATION_PUBLIC = this._config.APPLICATION_PUBLIC && this._config.APPLICATION_PUBLIC.toString() === 'true'
    this._config.CLEAR_QUEUE = config.CLEAR_QUEUE !== undefined ? config.CLEAR_QUEUE : true

    if (process.platform === 'win32') {
      this._config.CLEAR_QUEUE = false
    }

    if (this._config.APPLICATION_TYPE === 'worker') {
      this._app = new Worker(this._config)
    } else {
      this._app = new Fork(this._config)
    }

    this._app.sendError = this._sendError

    this._errors = {
      '-99': 'Ocurrio una excepcion, favor de intentar de nuevo.',
      '0': 'Smoke Test',
      '1': 'Este servicio no puede ser consumido de forma externa'
    }

    if (this._config.APPLICATION_TYPE === 'worker') {
      this._app.setErrors(this._errors)
    }

    if (this._config.MS_REDIS) {
      this.redisClient = redis.createClient({ host: this._config.MS_REDIS })
      this.redisClient.on('ready', () => {
        console.log('Redis UniqueID Connected')
      })
    }
  }

  setErrors (errors) {
    let newErrros = Object.assign({}, this._errors)
    Object.keys(errors).forEach(key => {
      if (!newErrros[key]) {
        newErrros[key] = errors[key]
      }
    })

    this._app.setErrors(newErrros)
  }

  setListener (listener) {
    this._listener = true
    this._app.setListener(listener)
  }

  setBifurcacion (fork) {
    if (this._config.APPLICATION_TYPE !== 'worker') {
      this._fork = true
      this._app.setBifurcacion(fork)
    }
  }

  _sendError (service, transactionId, error) {
    let message = Buffer.from(JSON.stringify({
      id: transactionId,
      servicio: service,
      error
    })).toString('base64')

    let producer = this._producer
    if (!producer) {
      producer = this.producer
    }

    producer.send([{
      topic: 'Errores_criticos',
      messages: [message],
      key: transactionId
    }], (err, _) => {
      if (err) console.error(err)
    })
  }

  _response (data, errorCode) {
    let errorResponse = {
      meta: {
        status: errorCode.toString() === '0' ? 'SUCCESS' : 'ERROR',
        id_transaction: data.metadata.id_transaction
      },
      data: {
        response: {
          errorCode,
          userMessage: this._errors[ errorCode.toString() ]
        }
      }
    }
    if (errorCode.toString() === '0') {
      errorResponse.data.response.hostname = this._hostname
      errorResponse.data.response.code = errorCode
      delete errorResponse.data.response.errorCode
    }

    data.response = errorResponse

    data.metadata.time = (new Date()).toISOString()
    data.metadata.worker = data.metadata.owner
    data.metadata.owner = this._config.TOPIC
    data.metadata.mtype = 'output'

    let uowner = ''
    let uworker = ''
    if ('uowner' in data.metadata) {
      uowner = data.metadata.uowner
    }

    if ('uworker' in data.metadata) {
      uworker = data.metadata.uworker
    }

    data.metadata.uworker = uowner
    data.metadata.uowner = uworker

    let topic = `respuesta_${data.metadata.owner}`

    data = {
      _id: data.metadata.id_transaction,
      response: data.response,
      metadata: data.metadata
    }

    if (this._config.DEBUG) {
      console.log(`SALIDA [${topic}]: ${JSON.stringify(data)}`)
    }

    data = Buffer.from(JSON.stringify(data)).toString('base64')

    this._producer.send([{
      topic,
      messages: [data]
    }], (err, _) => {
      if (err) console.error(err)
    })
  }

  startHttpHealthCheck(){
      this._appHealthCheck  = express()
      this._appHealthCheck.get('/', (req, res) => {
        if(this._producer) {

          const operationID = new Date().getTime().toString()
          let data = {
            data: { smoketest: true },
            headers: { Asynchronous: false, Authorizacion: '' },
            metadata: {
              intents: 0,
              callback: this._config.TOPIC,
              owner: this._config.TOPIC,
              uowner: this._config.TOPIC,
              uworker: `${this._config.TOPIC}_${operationID}`,
              worker: this._config.TOPIC,
              asynchronous: false,
              mtype: 'input',
              bifurcacion: false,
              from: this._hostname,

              id_operacion: operationID,
              id_transaction: uuidv4(),
              time: (new Date()).toISOString()
            }
          }

          let topic =  this._config.TOPIC
          data = Buffer.from(JSON.stringify(data)).toString('base64')
          this._producer.send([{
            topic,
            messages: [data]
          }], (err, _) => {
            if (err){
              console.error("HealthCheck] Error: "+ err)
              console.error(err)
              res.status(500).json({ error: err.toString() })
            }else{
              res.json({})
            } 
          })

        } else {
          console.error("[HealthCheck] Error: Producer no initialized")
          res.status(500).json({ error: 'Producer not initialized' })
        }
      })

      this._appHealthCheck.listen(80, () => {
        console.log('[HealthCheck] listening on 0.0.0.0:80')
      })
  }

  validateId (base64) {
    const md5 = crypto.createHash('md5').update(base64).digest("hex")
    const sha = crypto.createHash('sha256').update(base64).digest('hex')

    const uniqueID = `${md5}_${sha}`

    return new Promise((resolve, reject) => {
      this.redisClient.get(uniqueID, (err, reply) => {
        if (err) {
          reject(err)
        }
        if (reply) {
          resolve(true)
        } else {
          this.redisClient.set(uniqueID, '1', 'EX', 60 * 5)
          resolve(false)
        }
      })
    })
  }

  start () {
    if (!this._listener) {
      console.error('Listener not configured')
      return
    } else if (this._producer) {
      console.error('Microservice already started')
      return
    } else if (this._config.APPLICATION_TYPE !== 'worker' && !this._fork) {
      console.error('Bifurcacion not configured')
      return
    }

    
    this._client = new kafka.KafkaClient({ kafkaHost: this._config.KAFKA })
    this._producer = new kafka.HighLevelProducer(this._client)
    this.startHttpHealthCheck();

    this._producer.on('ready', () => {
      let options = {
        kafkaHost: this._config.KAFKA,
        groupId: this._config.TOPIC
      }

      if (this._consumer) {
        return
      }

      this._app.producer = this._producer

      let topics = [this._config.TOPIC]
      this._app.producer.client.loadMetadataForTopics([], function (error, results) {
        if (!error && results.length > 0) {
          let metadata = results.find(el => el.metadata)
          if (metadata) {
            metadata = metadata.metadata
            topics = Object.keys(metadata)
          }
        }
      });
      setInterval(() => {
        this._app.producer.client.refreshMetadata(topics, (error) => {
          if (error) console.log(error)
        })
      }, 30000)

      this._consumer = new kafka.ConsumerGroup(options, [this._config.TOPIC])
      this._consumer.on('connect', () => {
        console.log(`Conectado a kafka [${this._config.KAFKA} - ${this._config.TOPIC}`)
      })
      this._consumer.on('message', (message) => {
        const base64 = message.value.toString()
        const run = (base64) => {
          try {
            const data = JSON.parse(Buffer.from(base64, 'base64').toString('utf8'))

            if (data.data.smoketest) {
              return this._response(data, 0)
            }

            if (this._config.DEBUG && this._config.APPLICATION_TYPE === 'worker') {
              console.log(`ENTRADA [${this._config.TOPIC}]: HEADERS: [${JSON.stringify(data.headers)}] DATA: [${JSON.stringify(data.data)}]`)
            }

            if (data.metadata.callback === this._config.TOPIC && !this._config.APPLICATION_PUBLIC) {
              return this._response(data, 1)
            } else if (data.metadata.time.indexOf('Z') > -1 && this._config.CLEAR_QUEUE) {
              const date = new Date(data.metadata.time)
              const currDate = new Date()
              let diff = Math.round((currDate.getTime() - date.getTime()) / 60000)
              if (diff >= 5) {
                return
              }
            }

            this._app.message(data)
          } catch (err) {
            console.error(err)
            console.error(`ENTRADA [${this._config.TOPIC}]: [${message.value.toString()}]`)
            const transactionId = (new Date().getTime())
            this._sendError(this._config.TOPIC, transactionId, err)
            this._response({
              metadata: {
                owner: this._config.TOPIC,
                id_transaction: `error_${transactionId}`
              },
              headers: {},
              data: {}
            }, -99)
          }
        }

        if (this._config.MS_REDIS) {
          this.validateId(base64).then(exists => {
            if (!exists) {
              run(base64)
            }
          }).catch(err => {
            console.error(err)
            console.error(`ENTRADA [${this._config.TOPIC}]: [${message.value.toString()}]`)
            const transactionId = (new Date().getTime())
            this._sendError(this._config.TOPIC, transactionId, err)
            this._response({
              metadata: {
                owner: this._config.TOPIC,
                id_transaction: `error_${transactionId}`
              },
              headers: {},
              data: {}
            }, -99)
          })
        } else {
          run(base64)
        }
      })
      this._consumer.on('error', (err) => {
        console.error(err)
      })
    })

    this._producer.on('error', (err) => {
      console.error(err)
    })
  }
}

export default Coppel
